package PhotoFolder;
use strict;
use utf8;

=encoding utf-8

=head1 NAME

PhotoFolder - tool for publishing your personal folders with photos as web photo galleries

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use PhotoFolder;

=head1 DESCRIPTION

The tool for publishing your personal folders with photos as web photo galleries

=over 8

=item startup

Internal Mojolicious method. For internal use only

=back

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 SEE ALSO

L<Mojolicious>, L<CTK>, L<Image::Magick>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

Copyright 2021 Serż Minus (Sergey Lepenkov)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

See C<LICENSE> file or L<https://www.apache.org/licenses/LICENSE-2.0.txt>

=cut

use vars qw/ $VERSION /;
$VERSION = "1.00";

use Mojo::Base 'Mojolicious';
use Mojo::File qw/path/;

use File::Basename qw//;

use CTK::Util qw/preparedir sharedir sharedstatedir dtf/;
use CTK::ConfGenUtil;

use constant {
        PREFIX              => 'photofolder',
        DEBUGMODE           => 0,
    };

has 'documentroot';
has 'homedir';
has 'datadir';
has 'uriprefix';
has 'debugmode';

# Route
sub startup {
    my $self = shift;

    # DebugMode
    $self->debugmode(DEBUGMODE ? 1 : $ENV{PHOTOFOLDER_DEBUG} ? 1 : 0) unless $self->debugmode;

    # Logging
    $self->log->level($self->debugmode ? "debug" : "warn"); #->path('./server.log');

    # HomeDir
    my $homedir = path(sharedir(), PREFIX)->to_string(); # /usr/share/photofolder
    $self->homedir($homedir) unless defined $self->homedir;

    # Switch to installable home directory
    $self->home(Mojo::Home->new($homedir));

    # DataDir
    my $datadir = path(sharedstatedir(), PREFIX)->to_string(); # /var/lib/photofolder
    $self->datadir($datadir);

    # DocumentRoot
    my $documentroot = $ENV{PHOTOFOLDER_ROOT} || $ENV{CONTEXT_DOCUMENT_ROOT} || $ENV{DOCUMENT_ROOT} || $self->homedir;
    $self->documentroot($documentroot) unless defined $self->documentroot;

    # URIprefix
    my $uriprefix = $self->uriprefix // $ENV{PHOTOFOLDER_PREFIX} || $ENV{PHOTOFOLDER_BASE} || $ENV{CONTEXT_PREFIX} || "";
    $uriprefix = "" if $uriprefix =~ /^\/+$/;
    $self->uriprefix($uriprefix);

    # Reset static dirs
    $self->static->paths()->[0] = $self->documentroot; #unshift @{$static->paths}, '/home/sri/themes/blue/public';
    $self->static->paths()->[1] = $datadir;

    # Add renderer path (templates)
    push @{$self->renderer->paths}, $self->homedir;
    $self->static->paths()->[2] = $self->homedir;

    # Routes
    $self->routes->get('/' => {template => 'index'})->to('alpha#root');
    $self->routes->get('/env')->to('alpha#env') if $self->debugmode;
    $self->routes->get('/info')->to('alpha#info') if $self->debugmode;;

    # Favicon
    delete $self->static->extra->{'favicon.ico'};

    return 1;
}

1;

# Controller
package PhotoFolder::Controller::Alpha;
use Mojo::Base 'Mojolicious::Controller';

use Encode qw//;

use Mojo::File qw/path/;
use Mojo::Util qw/decode/;
use Mojo::JSON qw/from_json/;

use File::Spec qw//;
use File::Copy qw/copy/;
use Image::Magick;
use CPAN::Meta::YAML;

use CTK::Util qw/preparedir sharedir sharedstatedir dtf/;
use CTK::ConfGenUtil;

use vars qw/ $VERSION /;
$VERSION = "1.00";

use constant {
        CONTENT_TYPE_JSON   => "application/json",
        CONTENT_TYPE_TXT    => "text/plain; charset=utf-8",
        DATETIME_GMT_FORMAT => "%YYYY-%MM-%DD %hh:%mm:%ss %G",
        THUMBNAILS_DIR      => ".thumbs",
        THUMBNAILS_GEOMETRY => "!128x128",
        SUPPORTED_EXTS      => qr/^(jpg|jpeg|gif|bmp|png|tiff)$/i,
        INDEX_FILES         => [qw/index.json index.yml/],
    };

# Action
sub root {
    my $self = shift;
    my $error = "";
    #$self->app->log->debug("Ooooooops!");

    # Path from user request
    my $req_path = _uripath_prepare($self->req->param("dir"));
    $self->stash(path => $req_path);

    # Check document root directory
    return $self->reply->exception(sprintf("Root directory not exists: %s", $self->app->documentroot))->rendered(400)
        unless (-d $self->app->documentroot) || (-l $self->app->documentroot);

    # Real path on disk ($self->app->documentroot + $req_path)
    my $path = ($req_path eq '/') ? $self->app->documentroot : path($self->app->documentroot, $req_path)->to_string;
    return $self->render(text => sprintf("Path not exists: %s", $req_path), status => 404)
        unless (-d $path) || (-l $path);

    # Check basename
    my $cur_path = File::Basename::basename($req_path);
    return $self->render(text => sprintf("Can't use system directory: %s", $req_path), status => 400)
        if $cur_path eq THUMBNAILS_DIR;

    # Load config
    my $conf = {};
    for (@{(INDEX_FILES)}) {
        my $cf = $_;
        my $file = path($path, $_)->to_string;
        next unless -f $file;
        my $cont = decode('UTF-8', path($path, $_)->slurp);
        next unless $cont;
        if ($cf =~ /json/) {
            $conf = eval { from_json $cont };
            if ($@) {
                $error = sprintf("Can't load data from file \"%s\": %s", $file, $@);
                last;
            }
            unless (ref $conf eq 'HASH') {
                $error = sprintf("Index file \"%s\" did not return a JSON object", $file);
                last;
            }
            $conf->{config_file} = $file;
        } elsif ($cf =~ /yml/) {
            my $yaml = eval { CPAN::Meta::YAML->read_string($cont) };
            if ($@) {
                $error = sprintf("Can't load data from file \"%s\": %s (%s)", $file, $@, CPAN::Meta::YAML->errstr || 'Unknown CPAN::Meta::YAML error');
                last;
            }
            $conf = $yaml->[0] if ref $yaml eq 'CPAN::Meta::YAML';
            unless (ref $conf eq 'HASH') {
                $error = sprintf("Index file \"%s\" did not return a YAML object", $file);
                last;
            }
            $conf->{config_file} = $file;
        }
    }
    $self->stash(conf => $conf);

    # Bread crumbs
    my $bcarr = length($req_path) > 1 ? path($req_path)->dirname->to_array : path($req_path)->to_array;
    my @breadcrumbs = ();
    my @tmp = ();
    foreach my $nm (@$bcarr) {
        next unless length $nm;
        push @tmp, $nm;
        push @breadcrumbs, {
            name => $nm,
            rel => _uripath_join(@tmp),
            act => 0,
        }
    }
    push @breadcrumbs, {
        name => $cur_path,
        rel => "",
        act => 1,
    } if length($cur_path) && $cur_path ne '/';
    $self->stash(breadcrumbs => \@breadcrumbs);

    # Thumbnail dir
    my $thumbnail_dir = $self->req->param("thumb")
        ? path($path, THUMBNAILS_DIR)->make_path->to_string
        : path($path, THUMBNAILS_DIR)->to_string;

    # Get files
    my (@files, @dirs);
    my $need_thumbnail = 0;
    foreach my $f (_scanfiles($path)) {
        # Get full path to file on disk
        my $fpath = path($path, $f); # Encode::encode(locale_fs => $f)
        my $abs_path = $fpath->to_string;

        # Get ext
        my $ext = $f =~ /.+\.([^.]+)$/ ? $1 : '';

        # Get type
        my $type = "unknown";
        if (-f $abs_path) {
            $type = "file";
            next unless $ext =~ SUPPORTED_EXTS;
        } elsif (-d $abs_path) {
            $type = "dir";
        }

        # Rel Path
        my $rel_path = $fpath->to_rel($self->app->documentroot); # ->to_string
        my $rel_path_arr = $rel_path->to_array;

        # Thumbnail
        my $thumb_abs = path($thumbnail_dir, $f);
        my $thumb_rel = $thumb_abs->to_rel($self->app->documentroot)->to_array;
        my $thumbnail = $type eq "file" ? -e $thumb_abs->to_string ? 1 : 0 : 1;

        # Create Thumbnail
        my $thumb_rel_str = _uripath_join(@$thumb_rel);
        if (!$thumbnail && $self->req->param("thumb") && ($type eq "file") && $need_thumbnail <= 10) {
            $self->app->log->debug(sprintf("Create thumbnail #%d: %s", $need_thumbnail, $thumb_rel_str));
            my $err = $self->resize($abs_path, $thumb_abs->to_string, THUMBNAILS_GEOMETRY);
            if ($err) {
                $error = sprintf("Can't Create thumbnail %s from %s: %s", $thumb_rel_str, $abs_path, $err);
                $self->app->log->error($error);
            } else {
                $thumbnail = 1 if -e $thumb_abs->to_string;
            }
        }
        $need_thumbnail += $thumbnail ? 0 : 1;

        # Record
        my %rec = (
            name        => $f,
            ext         => $ext,
            type        => $type,
            title       => value($conf => "manifest", $f) || $f,
            path        => $req_path,
            abs         => $abs_path, # in filesystem
            rel         => _uripath_join(@$rel_path_arr), # in URI
            thumbnail   => $thumbnail,
            thumb_abs   => $thumb_abs->to_string, # in filesystem
            thumb_rel   => $thumb_rel_str, # in URI
        );
        if ($type eq "dir") {
            push @dirs, {%rec};
        } else {
            push @files, {%rec};
        }
    }

    $self->stash(
        files => \@files,
        dirs => \@dirs,
        need_thumbnail => $need_thumbnail ? 1 : 0,
        error => $error,
    );
    return 1;
}
sub env {
    my $self = shift;
    $self->res->headers->content_type(CONTENT_TYPE_TXT);
    return $self->render(text => $self->dumper(\%ENV));
}
sub info {
    my $self = shift; # Mojolicious::Routes

    # Logging
    #$self->app->log->info("Catched! Info log mark") if $ctk->verbosemode();

    # Response
    $self->res->headers->content_type(CONTENT_TYPE_JSON);
    return $self->render(json => {
            ok => \1,
            message => "Ok",
            generated => dtf(DATETIME_GMT_FORMAT, time(), 1),

            $self->app->debugmode ? (
                homedir => $self->app->home(),
                datadir => $self->app->datadir,
                documentroot => $self->app->documentroot,
                uriprefix => $self->app->uriprefix,
                #tempdir => $ctk->tempdir,
                #logdir => $ctk->logdir,
                #logfile => $ctk->logfile,
                #foo => $self->req->param("foo"),
                paths => $self->app->static->paths(),
            ) : (),
        });
}

sub resize {
    my $self = shift; # Mojolicious::Routes
    my $file_in  = shift;
    my $file_out = shift;
    my $geometry = shift;

    my $crop = $geometry =~ s/\!//g;
    my ($wbox, $hbox) = (0, 0);
    ($wbox, $hbox) = ($1,$2) if $geometry =~ /^(\d+)x(\d+)$/i;

    #debug("\t\tГеометрия изображения         : \"$geometry\"");
    #debug("\t\tВписывать с обрезкой?         : ", $crop ? 'YES' : 'NO');

    # Получаем исходные значения WxHxMIME
    my $image = Image::Magick->new;
    $image->Read($file_in);
    my ($wsrc, $hsrc, $mime) = $image->Get('width','height','mime');
    #debug("\t\tИсходные данные изображения   : $wsrc x $hsrc ($mime)");

    # Какие размеры нужны?
    my ($w,$h) = $crop
        ? ($wbox, $hbox)
        : _to_WxH($wbox, $hbox, $wsrc, $hsrc);
    #debug("\t\tНужны размеры                 : $w x $h");

    # Нужно ли делать ресайс?
    my $res = ($w >= $wsrc and $h >= $hsrc) ? 0 : 1;
    #debug("\t\tНужно ли делать ресайс?       : ", $res ? 'YES' : 'NO');
    #debug("\t\tНужно просто скопировать?     : ", !$res ? 'YES' : 'NO');
    unless ($res) {
        # Не нужно уменьшать? - выходим!
        copy($file_in, $file_out) or return "Copy failed: $!"; # Просто копируем
        #return {width => $wsrc, height => $hsrc};
        return;
    }

    # До какого размера уменьшать
    my ($tow, $toh) = $crop
        ? _calc_resize_out($wsrc, $hsrc, $w, $h)
        : _calc_resize_in($wsrc, $hsrc, $w, $h);
    #debug("\t\tДо какого размера уменьшать   : $tow x $toh");

    # Уменьшаем
    $image->Set(quality => 100);
    $image->Resize(geometry => $tow.'x'.$toh."+0+0", filter => 'Lanczos');

    #::debug("\t\tНужно ли делать обрезание?    : ", $crop ? 'YES' : 'NO');
    if ($crop) {
        # Общитываем смещение для обрезки
        my ($x, $y) = (int(($tow - $w) / 2),int(($toh - $h) / 2));
        #::debug("\t\tРассчитанные смещения обрезки : $x x $y");

        # Обрезаем, если это нужно
        $image->Crop(width => $w, height => $h, 'x' => $x, 'y' => $y);
    }

    # Записывыем файл
    $image->Write(filename => $file_out, compression => 'None');

    #unlink($tfn);
    #return $crop ? {width => $w, height => $h} : {width => $tow, height => $toh};
    return;
}

sub _calc_resize_out {
    # Процедура рассчета до какого предела можно уменьшить фотку для обрезания
    my $w  = shift; # Картинка SRC
    my $h  = shift;
    my $mw = shift; # Какой нужен итог
    my $mh = shift;

    if ($w/$mw > $h/$mh) {
        # Соотношение ширин больше соотношения высот -
        # ширину приводим к требуемой, а высоту рассчитываем
        return (int($w*$mh/$h), $mh);
    } else {
        # Наоборот
        return ($mw, int($h*$mw/$w));
    }
}
sub _calc_resize_in {
    # Процедура рассчета до какого предела можно уменьшить фотку без обрезания
    my $w  = shift; # Картинка SRC
    my $h  = shift;
    my $mw = shift; # Какой нужен итог
    my $mh = shift;

    if ($w/$mw < $h/$mh) {
        # Соотношение ширин меньше соотношения высот -
        # ширину приводим к требуемой, а высоту рассчитываем
        return (int($w*$mh/$h), $mh);
    } else {
        # Наоборот
        return ($mw, int($h*$mw/$w));
    }
}
sub _to_WxH {
    # процедура определяет ориентацию фотографии
    my $wmx = shift || 0; # требуется сделать таких размеров
    my $hmx = shift || 0;
    my $w = shift || $wmx; # Исходные данные
    my $h = shift || $hmx;

    return ($hmx,$wmx) if ($w > $h); # Фотка горизонтальная - ширина больше высоты
    return ($wmx,$hmx);              # Фотка вертикальная - высота больше ширины
}
sub _scanfiles {
    my $dir = shift // '/';
    my @files;

    # Read dir
    opendir(my $dh, $dir) or return @files;
        @files = readdir($dh);
    closedir($dh);

    # Prepare output
    my @out = ();
    foreach my $inf (sort {$a cmp $b} @files) {
        next unless length $inf;
        #my $f = Encode::decode(locale_fs => $inf);
        my $f = $inf; # Encode::decode(locale_fs => $inf); # utf8::decode($f);
        utf8::decode($f);
        next if $f eq THUMBNAILS_DIR;
        next if $f eq '.';
        next if $f eq '..';
        next if $f =~ /^\./;
        push @out, $f;
    }
    return @out;
}
sub _uripath_join {
    return join("/", "", @_) || "/";
}
sub _uripath_prepare {
    my $p = shift // "";
    my $rp = sprintf("/%s", $p); # Forcing adding leading slash (/)
       $rp =~ s/\.\.//g; # Removing "up dir" signature (..)
       $rp =~ s/\/{2,}/\//g; # Removing unnecessary slashes (//////->/)
       $rp =~ s/\/$//g; # Removing terminating slash (/$)
       $rp = '/' unless length($rp); # Default!
    return $rp;
}

1;

__END__
